//-----------------------------------------------------------------------------
// jondahli
//
// Dahli.java
//
// Core class for the Dahli language engine
//
// (c) 2011 Brandon Valosek
//-----------------------------------------------------------------------------

package com.bvalosek.jondahli.core;

// imports
import android.util.Log;
import java.util.List;
import java.io.InputStream;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.util.Map;
import java.util.HashMap;
import java.util.ArrayList;
import java.util.Random;
import java.util.Set;
import java.util.HashSet;

/** main class for the Dahli engine */
public class Dahli {
   
   // map of words to their object
   private Map<String, Word> mWords = new HashMap<String, Word>();

   // all the damn phrases
   private Map<String,Phrase> mPhrases = new HashMap<String, Phrase> ();

   // convo thread
   private List<ConvoEntry> mThread = new ArrayList<ConvoEntry> ();

   /** expose the convo thread */
   public List<ConvoEntry> getThread () {
      return mThread;
   }

   /** get known phrase count */
   public int getPhraseCount () {
   	return mPhrases.size ();
   }

   /** known words */
   public int getWordCount () {
      return mWords.size ();
   }

   /** get a response from the engine */
   public void doResponse (String sResponse) {

      // form phrase from the response
      Phrase r = stringToPhrase (sResponse);

      // if this human input is a response to something jon said
      ConvoEntry last = getLastResponse ();
      if (last != null) {
         processPair (last.mPhrase, r);
      }

      // create new convo entry
      mThread.add(new ConvoEntry(sResponse, 0, r));


      /* get all phrases matching what the user typed and create an
       * array of their corresponding scores
       */
      Set<Phrase> matchingPhrases = getSimilarPhrases (r);
      int[] scores = new int[matchingPhrases.size()];
      int totalScore = 0;
      int n = 0;
      for (Phrase p : matchingPhrases) {
         scores[n] = r.getScore(p);
         totalScore += scores[n];

         log ("(" + scores[n] + ")phrase= " + p);
         n++;
      }
      log ("totalScore=" + totalScore + " size(matchingPhrases)=" + matchingPhrases.size() );

      // randomly choose a matched phrase weighted by score
      int tot = 0;
      n = 0;
      Phrase matchedPhrase = null;
      Random rnd = new Random ();
      int thres = rnd.nextInt (totalScore );
      for (Phrase p : matchingPhrases) {
         tot += scores[n++]; 
         matchedPhrase = p;
         if (tot >= thres) {
            break;
         }
      }

      log ("chose \"" + matchedPhrase + "\" with a score of " +  scores[n-1] );

      // chose a random response from the matched phrase
      List<Phrase> responses = matchedPhrase.getLinks ();
      Phrase jonResponse = null;
      
      log ("potential responses = " +  responses.size() );

      if (responses.size() > 0) {
         jonResponse = responses.get ( rnd.nextInt (responses.size() ) );
      } else {
      	jonResponse = matchedPhrase;
      }

      // form response entry
      mThread.add( new ConvoEntry(jonResponse.getText() ,1,jonResponse) );
      
      // break up debug
      log ("--end doResponse()");
   }

   /** find all related phrases */
   private Set<Phrase> getSimilarPhrases (Phrase p) {
      Set<Phrase> ret = new HashSet<Phrase> ();

      // loop over words and add each phrase
      for (Word w : p.getWords() ) {
         ret.addAll ( w.getPhrases() );
      }

      // TODO: score each response?

      // return all matching phrases
      return ret;
   }

   /** convert a string into a list of Word objects */
   private Phrase  stringToPhrase (String s) {
      List<Word> wordList = new ArrayList<Word> ();
     
      // normalize the string first
      s = s.toLowerCase ();

      // see if we already have a phrase with this string
      if (mPhrases.containsKey(s) ) {
         return mPhrases.get (s);
      }
      
      // iterator over words
      String[] words = s.split(" ");
      for (String word : words) {
         Word w;
         // either create a fresh word or pull out existing
         if (mWords.containsKey (word)) {
            w = mWords.get(word);
         } else {
         	w = new Word(word);
         	mWords.put (word, w);
         }
         
       	// add the word the the list
       	wordList.add (w);
       }

      // create a new phrase with the wordlist we finished and add it
      // to the phrase map
      Phrase p = new Phrase(wordList);
      mPhrases.put (s, p);
     
      // add a link to each word back to this phrase
      for (Word w : wordList) {
         w.addPhrase (p);
      }

      // done
      return p;
   }

   /** get the previously said thing by jon */
   private ConvoEntry getLastResponse () {
   	if ( !mThread.isEmpty() ) {
   		return mThread.get (mThread.size() -1);
   	} else {
   		return null;
      }
   }

   /** process an input / response pair */
   public void processPair (Phrase in, Phrase re) {
      // link the input to the response and process the human phrase
      in.addLink (re);
      //log ("new link: " + in + "->" + re);
   }

   /** read in lines from an input stream to load up convo */
   public boolean loadConvoFromStream (InputStream is) {

      // wrap it in a buffered reader
      BufferedReader br = new BufferedReader ( new InputStreamReader(is) );

      // read in the lines
      String str, rsp;
      try {
         while ((str = br.readLine()) != null) {
            // read input/response pairs, convert to phrase, and process
            rsp = br.readLine();
            processPair (stringToPhrase(str), stringToPhrase(rsp) );
         }
      } catch (Exception e) {
         Log.e("jond","error=" + e);
         return false;
      }

      // made it
      return true;
   }
  
   /** log using the android Log class */
   private void log (String s) {
   	Log.d ("jond", s);
   }

}
